import abc
import numpy as np


class Geometry(abc.ABC):
    def __init__(self, time_dependent):
        self.time_dependent = time_dependent
        self.points = []
        self.time_points = []
        self.boundary_points = dict()
        self.boundary_points["num"] = None

    # @abc.abstractmethod
    def is_internal(self, x):
        """
        Returns True if the point x is internal to the Geometry.
        """
        pass

    # @abc.abstractmethod
    def is_boundary(self, x):
        """
        Returns True if the point x is on the boundary of the Geometry.
        """
        pass

    # @abc.abstractmethod
    def grid_points(self, n):
        """
        Returns a list of n grid points.
        """
        pass

    # @abc.abstractmethod
    def grid_points_on_boundary(self, n):
        """
        Returns a list of n grid points on the boundary.
        """
        pass

    # @abc.abstractmethod
    def random_points(self, n):
        """
        Returns a list of n random points.
        """
        pass

    # @abc.abstractmethod
    def random_points_on_boundary(self, n):
        """
        Returns a list of n random points on the boundary.
        """
        pass

    def uniform_time_dependent(self, n, time_start, time_end):
        """
        Returns a list of n time-dependent points.
        """
        pass

    def union(self, other):
        """
        Returns the union of the two geometries.
        """
        pass

    def intersection(self, other):
        """
        Returns the intersection of the two geometries.
        """
        pass

    def difference(self, other):
        # produce a list of points that are in self but not in other
        for boundary in self.boundary_points:

            self.boundary_points[boundary] = self.boundary_points[boundary][
                np.logical_not(other.is_internal(self.boundary_points[boundary]))
            ]

        # produce a list of points that are in other but not in self
        for boundary in other.boundary_points:

            self.boundary_points[num] = other.boundary_points[boundary][
                self.is_internal(other.boundary_points[boundary])
            ]

        # produce a list of points that are in both self and other
        self.points = self.points[np.logical_not(other.is_internal(self.points))]



from .geometry import Geometry
from .point import Point
from .polygon import Polygon
from .transform import Transform
from .box import Box
from .cylinder import Cylinder

__all__ = [
    "Geometry",
    "Point",
    "Polygon",
    "Sampler",
    "Transform",
    "Box",
    "Cylinder",
]



from .geometry import Geometry
from .sampler import Sampler
import numpy as np
from shapely.geometry import Polygon as ShapelyPolygon
from shapely.geometry import Point as ShapelyPoint


class Polygon(Geometry):
    def __init__(self, pointlist, time_dependent=False):
        super().__init__(time_dependent)
        self.shape = ShapelyPolygon(pointlist)
        self.pointlist = pointlist
        self.ndim = len(pointlist[0])
        self.n = len(pointlist)
        self.l_bounds = np.zeros(self.ndim)
        self.u_bounds = np.zeros(self.ndim)
        for i in range(self.ndim):
            self.l_bounds[i] = min([point[i] for point in pointlist])
            self.u_bounds[i] = max([point[i] for point in pointlist])

    def __str__(self):
        return "Polygon(%s)" % self.pointlist

    def _is_internal(self, x):
        return self.shape.intersects(ShapelyPoint(x))

    def is_internal(self, x):
        return [self._is_internal(point) for point in x]

    def _is_boundary(self, x):
        return self.shape.touches(ShapelyPoint(x))

    def is_boundary(self, x):
        return [self._is_boundary(point) for point in x]

    def grid_points(self, n):
        vlist = []
        for i in range(self.ndim):
            vlist.append(
                Sampler(1, n, self.l_bounds[i], self.u_bounds[i], samplingtype="grid")
            )

        points = np.vstack(np.meshgrid(*vlist)).reshape(self.ndim, -1).T
        points = points[self.is_internal(points)]
        self.points = points
        return points

    def grid_points_on_boundary(self, width):
        vlist = []
        for i in range(self.n):
            dx = self.pointlist[(i + 1) % self.n][0] - self.pointlist[i][0]
            dy = self.pointlist[(i + 1) % self.n][1] - self.pointlist[i][1]

            if dx == 0:
                y = Sampler(
                    1,
                    (1 / width, -1 / width)[dy < 0],
                    self.pointlist[i][1],
                    self.pointlist[(i + 1) % self.n][1],
                    samplingtype="grid",
                )
                x = np.full_like(y, self.pointlist[i][0])
            elif dy == 0:
                x = Sampler(
                    1,
                    (1 / width, -1 / width)[dx < 0],
                    self.pointlist[i][0],
                    self.pointlist[(i + 1) % self.n][0],
                    samplingtype="grid",
                )
                y = np.full_like(x, self.pointlist[i][1])
            else:
                d = np.sqrt(np.square(1 / width) / (1 + np.square(dx / dy)))
                x = Sampler(
                    1,
                    (d, -d)[dx < 0],
                    self.pointlist[i][0],
                    self.pointlist[(i + 1) % self.n][0],
                    samplingtype="grid",
                )
                y = (dy / dx) * (x - self.pointlist[i][0]) + self.pointlist[i][1]
            vlist.append(np.vstack((x, y)).T)
            # every segment will save like this: {'b0': [[x1, y1],...], 'b1': [[x2, y2],...], ...}
            self.boundary_points["b" + str(i)] = np.vstack((x, y)).T

        self.boundary_points["num"] = self.n
        # for development
        points = np.vstack(vlist)
        return points

    def random_points(self, n, samplingtype="random"):
        vlist = []
        for i in range(n):
            x = Sampler(
                1, 1, self.l_bounds[0], self.u_bounds[0], samplingtype=samplingtype
            )
            y = Sampler(
                1, 1, self.l_bounds[1], self.u_bounds[1], samplingtype=samplingtype
            )
            while not self._is_internal([x, y]):
                x = Sampler(
                    1, 1, self.l_bounds[0], self.u_bounds[0], samplingtype=samplingtype
                )
                y = Sampler(
                    1, 1, self.l_bounds[1], self.u_bounds[1], samplingtype=samplingtype
                )
            vlist.append([x, y])

        self.points = np.vstack(np.array(vlist).T).T.squeeze()
        return self.points

    def random_points_on_boundary(self, width):
        vlist = []
        for i in range(self.n):
            dx = self.pointlist[(i + 1) % self.n][0] - self.pointlist[i][0]
            dy = self.pointlist[(i + 1) % self.n][1] - self.pointlist[i][1]
            len = np.sqrt(dx ** 2 + dy ** 2)
            n = int(width * len)
            if dx == 0:
                y = Sampler(
                    1,
                    n,
                    self.pointlist[i][1],
                    self.pointlist[(i + 1) % self.n][1],
                    samplingtype="random",
                )
                x = np.full_like(y, self.pointlist[i][0])
            elif dy == 0:
                x = Sampler(
                    1,
                    n,
                    self.pointlist[i][0],
                    self.pointlist[(i + 1) % self.n][0],
                    samplingtype="random",
                )
                y = np.full_like(x, self.pointlist[i][1])
            else:
                x = Sampler(
                    1,
                    n,
                    self.pointlist[i][0],
                    self.pointlist[(i + 1) % self.n][0],
                    samplingtype="random",
                )
                y = (dy / dx) * (x - self.pointlist[i][0]) + self.pointlist[i][1]
            vlist.append(np.vstack((x, y)).T)
            # every segment will save like this: {'b0': [[x1, y1],...], 'b1': [[x2, y2],...], ...}
            self.boundary_points["b" + str(i)] = np.vstack((x, y)).T

        self.boundary_points["num"] = self.n
        # for development
        points = np.vstack(vlist)
        return points



from .geometry import Geometry
import numpy as np
from .box import Box


class Cylinder(Geometry):
    def __init__(
        self, pos, radius=1, length=1, q: list = [0, 0, 0], time_dependent=False
    ):
        """
        Creat a Cylinder.
        ### Args:
        - `pos` : The position of the Cylinder.
        - `radius` : The radius of the Cylinder. Default is 1.
        - `length` : The length of the Cylinder. Default is 1.
        - `q` : The euler angles of the Cylinder. Default is [0, 0, 0].\n
        First angle is the rotation around the z-axis, second angle is the\n
        rotation around the y-axis, and the third angle is the rotation around\n
        the x-axis.
        - `time_dependent` : True if the Cylinder is time dependent. Default is False.

        ### Returns:
        - `Cylinder` : A Cylinder.
        """
        super().__init__(time_dependent)
        self.pos = pos
        self.radius = radius
        self.length = length
        self.q = q
        self.transformationMatrix = Box.get_transformationMatrix(self.q)

    def __str__(self):
        return "Cylinder({}, {}, {})".format(self.pos, self.radius, self.axis)

    def is_internal(self, x):
        """
        Returns True if the point x is internal to the Geometry.
        ### Args:
        - `x` : A point as like [x, y, z]
        ### Returns:
        - `bool` : True if the point x is internal to the Geometry.
        """
        x = np.array(x)
        x = np.dot(np.linalg.inv(self.transformationMatrix), x.T).T
        x = x - self.pos
        print(x.shape)
        r_log = np.linalg.norm(x[:, :-1], axis=1) < self.radius
        l_log = np.logical_and(x[:, -1] < self.length, x[:, -1] > 0)
        return np.logical_and(r_log, l_log)

    def _is_boundary(self, x):
        """
        Returns True if the point x is on the boundary of the Geometry.
        ### Args:
        - `x` : A point as like [x, y, z]
        ### Returns:
        - `bool` : True if the point x is on the boundary of the Geometry.
        """
        x = np.array(x)
        x = np.dot(np.linalg.inv(self.transformationMatrix), x.T).T
        x = x - self.pos
        return np.linalg.norm(x) == self.radius

    def is_boundary(self, x):
        """
        Returns True if the point x is on the boundary of the Geometry.
        ### Args:
        - `x` : list of points as like [[x, y, z], [x, y, z], ...]
        ### Returns:
        - `bool` : True if the point x is on the boundary of the Geometry.
        """
        if isinstance(x, list):
            return [self._is_boundary(xi) for xi in x]
        else:
            return self._is_boundary(x)

    def grid_points(self, n):
        """
        Returns a list of points on the Geometry.
        ### Args:
        - `n` : Number of points on the Geometry.
        ### Returns:
        - `list` : A list of points on the Geometry.
        """
        x = np.linspace(-self.radius, self.radius, n)
        y = np.linspace(-self.radius, self.radius, n)
        z = np.linspace(0, self.length, n)
        points = np.meshgrid(x, y, z)
        points = np.array(points).reshape(3, -1).T
        points = np.dot(self.transformationMatrix, points.T).T
        point = points[self.is_internal(points)]
        point = point + self.pos
        points = np.dot(self.transformationMatrix, point.T).T
        return points

    def grid_points_on_boundary(self, n):
        # boundary_points = {}
        theta = np.linspace(0, 2 * np.pi, n)
        l_po = np.linspace(0, self.length, n)
        vlist = []
        for i in range(n):
            point = np.array(
                [
                    self.radius * np.cos(theta),
                    self.radius * np.sin(theta),
                    l_po[i] * np.ones(n),
                ]
            )
            print(point.shape)
            vlist.append(point.reshape(-1, 3))

        vlist = np.vstack(vlist)
        vlist = vlist + self.pos
        vlist = np.dot(self.transformationMatrix, vlist.T).T
        return vlist



from .geometry import Geometry
import numpy as np


class Box(Geometry):
    def __init__(
            self,
            center: list,
            length: float,
            width: float,
            height: float,
            q: list = [0, 0, 0],
            time_dependent=False,
    ) -> None:
        super().__init__(time_dependent)
        self.center = center
        self.q = q
        self.length = length
        self.width = width
        self.height = height
        self.transformationMatrix = self.get_transformationMatrix()
        self.above_points = self.get_above_points()
        self.below_points = self.get_below_points()

    def __str__(self):
        return "Box: center: {}, q: {}, length: {}, width: {}, height: {}, ".format(
            self.center, self.q, self.length, self.width, self.height
        )

    @staticmethod
    def get_transformationMatrix(q):
        """
        Calculate rotation matrix from Euler angle meter.

        We first rotate around the z axis, then around the y axis,
        and finally around the x axis.

        ### TODO: check if q is correct
        ### Return:
        - `transformationMatrix`: transformation matrix
        """

        # compute the transformation matrix
        # we use the euler angles to rotate the box z-y-x
        gamma = np.array(
            [
                [np.cos(q[0]), -np.sin(q[0]), 0],
                [np.sin(q[0]), np.cos(q[0]), 0],
                [0, 0, 1],
            ]
        )
        beta = np.array(
            [
                [1, 0, 0],
                [0, np.cos(q[1]), -np.sin(q[1])],
                [0, np.sin(q[1]), np.cos(q[1])],
            ]
        )
        alpha = np.array(
            [
                [np.cos(q[2]), 0, np.sin(q[2])],
                [0, 1, 0],
                [-np.sin(q[2]), 0, np.cos(q[2])],
            ]
        )
        return np.dot(np.dot(gamma, beta), alpha)

    def do_transform(self, x):
        """
        Transform a point from the box to the global frame.

        ### Args:
        - `x` (dict): point to transform
        ### Return:
        - `x` (dict): transformed point
        """
        for key in x:
            x[key] = x[key] - self.center
            x[key] = np.dot(self.transformationMatrix, x[key].T).T
            x[key] = x[key] + self.center
        return x

    def get_above_points(self):
        above_point = np.array(
            [
                [
                    self.center[0] - self.length / 2,
                    self.center[1] - self.width / 2,
                    self.center[2] + self.height,
                ],
                [
                    self.center[0] + self.length / 2,
                    self.center[1] - self.width / 2,
                    self.center[2] + self.height,
                ],
                [
                    self.center[0] + self.length / 2,
                    self.center[1] + self.width / 2,
                    self.center[2] + self.height,
                ],
                [
                    self.center[0] - self.length / 2,
                    self.center[1] + self.width / 2,
                    self.center[2] + self.height,
                ],
            ]
        )
        return above_point

    def get_below_points(self):
        below_point = np.array(
            [
                [
                    self.center[0] - self.length / 2,
                    self.center[1] - self.width / 2,
                    self.center[2],
                ],
                [
                    self.center[0] + self.length / 2,
                    self.center[1] - self.width / 2,
                    self.center[2],
                ],
                [
                    self.center[0] + self.length / 2,
                    self.center[1] + self.width / 2,
                    self.center[2],
                ],
                [
                    self.center[0] - self.length / 2,
                    self.center[1] + self.width / 2,
                    self.center[2],
                ],
            ]
        )
        return below_point

    def _is_internal(self, x):
        """
        Check if a point is inside the box.

        ### Args:
        - `x` (list): point to check
        ### Return:
        - `is_internal` (bool): True if point is inside the box, False otherwise
        """
        # check if point is inside the box
        if np.all(x >= self.center - self.length / 2) and np.all(
                x <= self.center + self.length / 2
        ):
            return True
        else:
            return False

    def is_internal(self, x):
        """
        Check if a list point is inside the box.
        ### Args:
        - `x` (list): list of points to check like [[x1, y1, z1], [x2, y2, z2], ...]
        ### Return:
        - `is_internal` (list): True if point is inside the box, False otherwise
        """
        if isinstance(x, list):
            return [self._is_internal(i) for i in x]
        else:
            return self._is_internal(x)

    def _is_boundary(self, x):
        """
        Check if a point is on the boundary of the box.

        ### Args:
        - `x` (list): point to check
        ### Return:
        - `is_boundary` (bool): True if point is on the boundary of the box, False otherwise
        """
        # check if point is on the boundary of the box
        if np.all(x == self.center - self.length / 2) or np.all(
                x == self.center + self.length / 2
        ):
            return True
        else:
            return False

    def is_boundary(self, x):
        """
        Check if a list point is on the boundary of the box.
        ### Args:
        - `x` (list): list of points to check like [[x1, y1, z1], [x2, y2, z2], ...]
        ### Return:
        - `is_boundary` (list): True if point is on the boundary of the box, False otherwise
        """
        return [self._is_boundary(i) for i in x]

    def grid_points(self, n):
        """
        Generate a grid of points inside the box.

        ### Args:
        - `n` (int): spacing of the grid
        ### Return:
        - `grid_points` (list): list of points inside the box
        """
        # generate a list of points on the box
        x = np.linspace(
            self.center[0] - self.length / 2, self.center[0] + self.length / 2, n
        )
        y = np.linspace(
            self.center[1] - self.width / 2, self.center[1] + self.width / 2, n
        )
        z = np.linspace(self.center[2], self.center[2] + self.height, n)
        points = np.vstack(np.meshgrid(x, y, z)).reshape(3, -1).T
        points = np.dot(self.transformationMatrix, points.T).T
        self.points = points
        return points

    def grid_points_on_edge(self, n):
        """
        Generate a grid of points on the edge of the box.

        ### Args:
        - `n` (int): spacing of the grid
        ### Return:
        - `grid_points` (list): list of points on the edge of the box
        """
        edge_points = {}
        for i in [0, 2]:
            point = np.arange(
                self.above_points[i, 0],
                self.above_points[(i + 1) % 4, 0],
                1
                / n
                * (1, -1)[self.above_points[i, 0] > self.above_points[(i + 1) % 4, 0]],
            )
            edge_points[i] = np.vstack(
                (
                    point,
                    self.above_points[i, 1] * np.ones(len(point)),
                    self.above_points[i, 2] * np.ones(len(point)),
                )
            ).T

            point = np.arange(
                self.below_points[i, 0],
                self.below_points[(i + 1) % 4, 0],
                1
                / n
                * (1, -1)[self.below_points[i, 0] > self.below_points[(i + 1) % 4, 0]],
            )
            edge_points[i + 4] = np.vstack(
                (
                    point,
                    self.below_points[i, 1] * np.ones(len(point)),
                    self.below_points[i, 2] * np.ones(len(point)),
                )
            ).T

        for i in [1, 3]:
            point = np.arange(
                self.above_points[i, 1],
                self.above_points[(i + 1) % 4, 1],
                1
                / n
                * (1, -1)[self.above_points[i, 1] > self.above_points[(i + 1) % 4, 1]],
            )
            edge_points[i] = np.vstack(
                (
                    self.above_points[i, 0] * np.ones(len(point)),
                    point,
                    self.above_points[i, 2] * np.ones(len(point)),
                )
            ).T

            point = np.arange(
                self.below_points[i, 1],
                self.below_points[(i + 1) % 4, 1],
                1
                / n
                * (1, -1)[self.below_points[i, 1] > self.below_points[(i + 1) % 4, 1]],
            )
            edge_points[i + 4] = np.vstack(
                (
                    self.below_points[i, 0] * np.ones(len(point)),
                    point,
                    self.below_points[i, 2] * np.ones(len(point)),
                )
            ).T

        for i in range(4):
            point = np.arange(
                self.above_points[i, 2],
                self.below_points[i, 2],
                1 / n * (1, -1)[self.above_points[i, 2] > self.below_points[i, 2]],
            )
            edge_points[i + 8] = np.vstack(
                (
                    self.above_points[i, 0] * np.ones(len(point)),
                    self.above_points[i, 1] * np.ones(len(point)),
                    point,
                )
            ).T

        edge_points = self.do_transform(edge_points)

        return edge_points

    def grid_points_on_surface(self, spacing):
        """
        Generate a dict of points on the box surface.

        ### Args:
        - `spacing`: spacing between points on the surface

        ### Returns:
        - `surface_points`: dict of points on the box surface
        """

        surfacepoints = {}
        x = np.arange(
            self.center[0] - self.length / 2,
            self.center[0] + self.length / 2,
            1 / spacing * (1, -1)[self.center[0] > self.center[0] + self.length / 2],
        )
        y = np.arange(
            self.center[1] - self.width / 2,
            self.center[1] + self.width / 2,
            1 / spacing * (1, -1)[self.center[1] > self.center[1] + self.width / 2],
        )
        z = np.arange(
            self.center[2],
            self.center[2] + self.height,
            1 / spacing * (1, -1)[self.center[2] > self.center[2] + self.height],
        )
        flag = 1
        for i in range(2):
            point = (
                np.meshgrid(x, self.center[1] + flag * self.width / 2, z)
                    .reshape(3, -1)
                    .T
            )
            surfacepoints[i] = np.dot(self.transformationMatrix, point.T).T
            flag *= -1

        for i in range(2):
            point = (
                np.meshgrid(self.center[0] + flag * self.length / 2, y, z)
                    .reshape(3, -1)
                    .T
            )
            surfacepoints[i + 2] = np.dot(self.transformationMatrix, point.T).T
            flag *= -1

        point = np.meshgrid(x, y, self.center[2] + self.height).reshape(3, -1).T
        surfacepoints[4] = np.vstack((point[:, 0], point[:, 1], point[:, 2])).T
        point = np.meshgrid(x, y, self.center[2]).reshape(3, -1).T
        surfacepoints[5] = np.vstack((point[:, 0], point[:, 1], point[:, 2])).T

        return surfacepoints

    def grid_points_on_boundary(self, spacing):
        """
        Generate a list of points on the box boundary.

        ### Args:
        - `spacing`: spacing between points on the boundary

        ### Returns:
        - `boundary_points`: list of points on the box boundary
        """

        edge_points = self.grid_points_on_edge(spacing)
        surfacepoints = self.grid_points_on_surface(spacing)

        boundary_points = {}
        for i in range(12):
            boundary_points["e" + str(i)] = edge_points[i]
        for i in range(6):
            boundary_points["s" + str(i)] = surfacepoints[i]
        self.boundary_points = boundary_points
        return boundary_points

    def random_points(self, n):
        """
        Generate a dict of random points on the box.
        ### Args:
        - `n` : number of points
        ### Returns:
        dict: dict of random points on the box
        """
        vlist = []
        for i in range(n):
            x = np.random.uniform(
                self.center[0] - self.length / 2, self.center[0] + self.length / 2
            )
            y = np.random.uniform(
                self.center[1] - self.width / 2, self.center[1] + self.width / 2
            )
            z = np.random.uniform(self.center[2], self.center[2] + self.height)
            vlist.append([x, y, z])

        points = np.vstack(vlist)
        points = np.dot(self.transformationMatrix, points.T).T
        return points

    def random_points_on_edge(self, n):
        """
        Generate a dict of random points on the box edges.
        ### Args:
        n (int): number of points

        ### Returns:
        dict: dict of random points on the box edges
        """
        edge_points = {}
        # generate points towards the x-axis and the n is like self.length
        for i in [0, 2]:
            point = np.random.uniform(
                self.above_points[i, 0],
                self.above_points[(i + 1) % 4, 0],
                int(n * (self.length / self.length)),
            )
            edge_points[i] = np.vstack(
                (
                    point,
                    self.above_points[i, 1] * np.ones(len(point)),
                    self.above_points[i, 2] * np.ones(len(point)),
                )
            ).T
            point = np.random.uniform(
                self.below_points[i, 0],
                self.below_points[(i + 1) % 4, 0],
                int(n * (self.length / self.length)),
            )
            edge_points[i + 4] = np.vstack(
                (
                    point,
                    self.below_points[i, 1] * np.ones(len(point)),
                    self.below_points[i, 2] * np.ones(len(point)),
                )
            ).T

        # generate points towards the y-axis and the n is like self.width
        for i in [1, 3]:
            point = np.random.uniform(
                self.above_points[i, 1],
                self.above_points[(i + 1) % 4, 1],
                int(n * (self.width / self.length)),
            )
            edge_points[i] = np.vstack(
                (
                    self.above_points[i, 0] * np.ones(len(point)),
                    point,
                    self.above_points[i, 2] * np.ones(len(point)),
                )
            ).T

            point = np.random.uniform(
                self.below_points[i, 1],
                self.below_points[(i + 1) % 4, 1],
                int(n * (self.width / self.length)),
            )
            edge_points[i + 4] = np.vstack(
                (
                    self.below_points[i, 0] * np.ones(len(point)),
                    point,
                    self.below_points[i, 2] * np.ones(len(point)),
                )
            ).T

        # generate points towards the z-axis and the n is like self.height
        for i in range(4):
            point = np.random.uniform(
                self.above_points[i, 2],
                self.below_points[i, 2],
                int(n * (self.height / self.length)),
            )
            edge_points[i + 8] = np.vstack(
                (
                    self.above_points[i, 0] * np.ones(len(point)),
                    self.above_points[i, 1] * np.ones(len(point)),
                    point,
                )
            ).T

        edge_points = self.do_transform(edge_points)

        return edge_points

    def random_points_on_surface(self, n):
        """
        Generate a dict of points on the box surface.

        ### Args:
        n (int): number of points

        ### Returns:
        dict of points on the box surface
        """
        surfacepoints = {}
        for i in range(2):
            point = []
            for j in range(int(n)):
                x = np.random.uniform(
                    self.center[0] - self.length / 2, self.center[0] + self.length / 2
                )
                y = np.random.uniform(
                    self.center[1] - self.width / 2, self.center[1] + self.width / 2
                )
                point.append([x, y, self.center[2] + self.height * i])
            surfacepoints[i] = np.vstack(point)

        for i in range(2):
            point = []
            for j in range(int(n * (self.height / self.width))):
                x = np.random.uniform(
                    self.center[0] - self.length / 2, self.center[0] + self.length / 2
                )
                z = np.random.uniform(self.center[2], self.center[2] + self.height)
                point.append(
                    [x, self.center[1] + (self.width / 2) * (1, -1)[i == 0], z]
                )
            surfacepoints[i + 2] = np.vstack(point)

        for i in range(2):
            point = []
            for j in range(int(n * (self.height / self.length))):
                y = np.random.uniform(
                    self.center[1] - self.width / 2, self.center[1] + self.width / 2
                )
                z = np.random.uniform(self.center[2], self.center[2] + self.height)
                point.append(
                    [self.center[0] + (self.length / 2) * (1, -1)[i == 0], y, z]
                )
            surfacepoints[i + 4] = np.vstack(point)

        surfacepoints = self.do_transform(surfacepoints)

        return surfacepoints

    def random_points_on_boundary(self, surfacen: int, edgen: int = None) -> dict:
        """
        Generate a dict of points on the box boundary.
        ### Args:
        - surfacen (int): number of points on the surface
        - edgen (int): number of points on the edge (default: None)

        ### Note:
        - If edgen is None, surfacen square will be set as edgen.

        ### Returns:
        dict: dict of points on the box boundary
        """
        if edgen is None:
            edgen = int(np.floor(np.sqrt(surfacen)))
        boundary = {}
        edge_points = self.random_points_on_edge(edgen)
        surface_points = self.random_points_on_surface(surfacen)

        for i in range(6):
            boundary["s" + str(i)] = surface_points[i]

        for i in range(12):
            boundary["e" + str(i)] = edge_points[i]

        self.boundary_points = boundary

        return boundary



import numpy as np


class Transform:
    @staticmethod
    def rotate(p, angle):
        """
        Rotate a point around the origin by a given angle.
        p: pointList to rotate
        angle: angle to rotate by
        """
        print(p.shape)
        c, s = np.cos(angle), np.sin(angle)
        R = np.array(((c, -s), (s, c)))
        return np.dot(p, R)

    @staticmethod
    def translate(p, x, y):
        """
        Translate a point by a given amount.
        p: pointList to translate
        x: amount to translate in the x direction
        y: amount to translate in the y direction
        """
        return np.add(p, np.array((x, y)))

    @staticmethod
    def scale(p, x, y):
        """
        Scale a point by a given amount.
        p: pointList to scale
        x: amount to scale in the x direction
        y: amount to scale in the y direction
        """
        return np.multiply(p, np.array((x, y)))

    @staticmethod
    def expand(p, theta):
        """
        Expand a point by a given angle.
        p: pointList to expand
        theta: angle to expand by
        """
        r = p[:, 0]
        x = np.array([r * np.cos(theta), r * np.sin(theta), p[:, 1]]).T
        return x



from .sampler import Sampler
from .geometry import Geometry
import numpy as np


class Point(Geometry):
    def __init__(self, center, radius, time_dependent=False):
        super().__init__(time_dependent)
        self.center = center
        self.radius = radius
        self.dim = len(center)

    def __str__(self):
        return "Point({}, {})".format(self.center, self.radius)

    def is_internal(self, x):
        return np.linalg.norm(x - self.center, axis=1) <= self.radius

    def is_boundary(self, x):
        return np.linalg.norm(x - self.center, axis=1) == self.radius

    def grid_points(self, n):
        vlist = []
        for i in range(self.dim):
            vlist.append(
                Sampler(
                    1,
                    n,
                    self.center[i] - self.radius,
                    self.center[i] + self.radius,
                    type="grid",
                )
            )

        points = np.vstack(np.meshgrid(*vlist)).reshape(self.dim, -1).T
        points = points[self.is_internal(points)]
        self.points = points
        return points

    def grid_points_on_boundary(self, n):
        if self.dim == 2:
            theta = Sampler(1, n, 0, 2 * np.pi, type="grid")
            points = np.vstack(
                (
                    self.center[0] + self.radius * np.cos(theta),
                    self.center[1] + self.radius * np.sin(theta),
                )
            ).T
        elif self.dim == 3:
            i = Sampler(1, n, 0, n, type="grid")
            phi = np.arccos(((2 * i - 1) / n - 1).clip(-1, 1))
            theta = np.sqrt(n * np.pi) * phi
            points = np.vstack(
                (
                    self.center[0] + self.radius * np.cos(theta) * np.sin(phi),
                    self.center[1] + self.radius * np.sin(theta) * np.sin(phi),
                    self.center[2] + self.radius * np.cos(phi),
                )
            ).T
        self.boundary_points["b0"] = points
        self.boundary_points["num"] = 1
        return points

    def random_points(self, n):
        # http://extremelearning.com.au/how-to-generate-uniformly-random-points-on-n-spheres-and-n-balls/
        u = Sampler(self.dim, n, 0, 1, samplingtype="normal")
        norm = np.sum(u ** 2, axis=0) ** (0.5)
        r = self.radius * np.random.random(n) ** (1.0 / self.dim)
        points = r * u / norm
        points = points.T + self.center
        self.points = points
        return points

    def random_points_on_boundary(self, n, seed=None):
        if self.dim == 2:
            theta = Sampler(1, n, 0, 2 * np.pi, samplingtype="random")
            points = np.vstack(
                (self.radius * np.cos(theta), self.radius * np.sin(theta))
            ).T
        elif self.dim == 3:
            u = Sampler(1, n, 0, 1, samplingtype="normal")
            v = Sampler(1, n, 0, 1, samplingtype="normal")
            w = Sampler(1, n, 0, 1, samplingtype="normal")

            norm = np.sqrt(u ** 2 + v ** 2 + w ** 2)
            x = u / norm
            y = v / norm
            z = w / norm

            points = np.vstack((x, y, z)).T
            points = points * self.radius + self.center
        self.boundary_points["b0"] = points
        self.boundary_points["num"] = 1
        return points



from setuptools import setup, find_packages
from os import path

packageName = "pygeomesh"

this_directory = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# version scheme: major.minor.patch
versionInfo = 0, 1, 1

__version__ = ".".join(map(str, versionInfo))


setup(
    name=packageName,
    version=__version__,
    author="PuQing",
    author_email="me@puqing.work",
    packages=find_packages(),
    description="PyGeomesh is a tool for generating discretized points from geometry.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AndPuQing/PyGeomesh",
    download_url="https://pypi.python.org/pypi/pygeomesh",
    license="Apache License 2.0",
    platforms="any",
    requires=[
        "numpy",
        "scipy",
        "matplotlib",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
)
